/*************************************************************************
 *
 * Author : Kelvin Lau
 * Date   : Aug 10, 1998
 *
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "xcon.h"

/*************************************************************************
 * Motif, Xt and X related header files
 *************************************************************************/

#include <Xm/MainW.h>
#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/Xm.h>
#include <Xm/ScrolledW.h>
#include <Xm/Protocols.h>
#include <Xm/LabelG.h>
#include <Xm/PushB.h>
#include <Xm/MwmUtil.h>
#include <Xm/DialogS.h>
#include <Xm/PanedW.h>

/*************************************************************************
 * Global variables
 *************************************************************************/

XTextProperty Xcutbuf;
XtAppContext  Xapp;
XEvent        Xevent;
XmString      xmLen;
String        Str;
Display       *dpy;
Window        window, wRoot, wChild;
Widget        wTop, wForm, wText[BOX_COUNT], wButton[BOX_COUNT], 
              wLabel[BOX_COUNT], wlLength, wPopup, wPopForm, wList;
Arg           stArg[20];
ulong         ulVal;
int           i, iArg, iPos, iItemCnt=0, iRx, iRy, iWx, iWy;
uint          uiMask;
short         sCurButton=0, sPre=0, sLOffset=2, sTOffset=0, sMHeight=5, 
              sMWidth=5, sSThick=2, sBoxWidth=BOX_WIDTH, 
              sLabelDistance=LABEL_DISTANCE;
char          *pcOut, caLName[][4]={"HEX", "DEC", "ABC"},
              caTempName[20], cFlag=0, caFont[10]="TIMES12";

String userfonts[] = {
  "xcon*wLabel*fontList:\
  -misc-fixed-medium-r-semicondensed-*-*-90-100-*-c-*-iso8859-*=FIXED9,\
  -adobe-times-medium-r-normal-*-*-80-*-*-p-*-iso8859-*=TIMES8,\
  -adobe-times-medium-r-normal-*-*-100-*-*-p-*-iso8859-*=TIMES10,\
  -adobe-times-bold-r-normal-*-*-120-*-*-p-*-iso8859-*=TIMES12", NULL
};

/*************************************************************************
 * Function prototypes
 *************************************************************************/

void vPushed(Widget widget, XtPointer client_data, XtPointer call_data);
void vABCToHEX(char *pcData, char **ppcOut);
void vDECToHEX(ulong ulData, char **ppcOut);
void vABCToDEC(char *pcData, char **ppcOut);
void vHEXToDEC(ulong ulData, char **ppcOut);
void vHEXToABC(char *pcData, char **ppcOut);
void vConvertInput(char *pcInput);
void vDoubleClick(Widget widget, XtPointer client_data, XtPointer call_data);
void vPopDown(Widget widget, XtPointer client_data, XtPointer call_data);
void vClose(void);



/*************************************************************************
 * Main
 *************************************************************************/

int main(int argc, char *argv[])
{
  if (!strcmp(argv[argc-1], "-gnome"))
  {
	argc--;
	strcpy(caFont, "FIXED9");
	sTOffset=-2;
	sMHeight=0;
	sMWidth=2;
	sSThick=1;
	sBoxWidth=G_BOX_WIDTH;
	sLabelDistance=G_LABEL_DISTANCE;
  }

  wTop = XtVaAppInitialize(&Xapp, "wTop", NULL, 0, &argc, argv, userfonts, 
						   NULL, XmNheight, XCON_HEIGHT, XmNwidth, XCON_WIDTH,
						   NULL);

  dpy = XtDisplay(wTop);
  window = DefaultRootWindow(dpy);

#if 0
  XSelectInput(dpy, window, PropertyChangeMask|LeaveWindowMask);
#else
  XSelectInput(dpy, window, PropertyChangeMask);
#endif

  wForm = XtVaCreateManagedWidget("wForm", xmFormWidgetClass, wTop, NULL);

  /*************************************************************************
   * Scroll text boxes and their labels
   *************************************************************************/
  iArg = 0;
  XtSetArg(stArg[iArg], XmNtopAttachment, XmATTACH_WIDGET); iArg++;
  XtSetArg(stArg[iArg], XmNtopWidget, wText); iArg++;
  XtSetArg(stArg[iArg], XmNleftAttachment, XmATTACH_WIDGET); iArg++;
  XtSetArg(stArg[iArg], XmNleftWidget, wForm); iArg++;
  XtSetArg(stArg[iArg], XmNwidth, sBoxWidth); iArg++;
  XtSetArg(stArg[iArg], XmNshadowThickness, sSThick); iArg++;
  XtSetArg(stArg[iArg], XmNmarginHeight, sMHeight); iArg++;
  XtSetArg(stArg[iArg], XmNmarginWidth, sMWidth); iArg++;
  XtSetArg(stArg[iArg], XmNtopOffset, sTOffset); iArg++;
  for (i = 0; i < BOX_COUNT; i++)
  {
	xmLen = XmStringCreate(caLName[i], caFont);
	wLabel[i] = XtVaCreateManagedWidget("wLabel", xmLabelGadgetClass, wForm,
										XmNlabelString, xmLen,
										XmNtopAttachment, XmATTACH_WIDGET,
										XmNtopWidget, wForm,
										XmNleftAttachment, XmATTACH_WIDGET,
										XmNleftWidget, wLabel[i-sPre],
										XmNtopOffset, TOPOFFSET,
										XmNleftOffset, sLOffset,
										XmNmarginHeight, 0,
										NULL);

	XtSetArg(stArg[1], XmNtopWidget, wLabel[i]);
	wText[i] = XmCreateScrolledText(wForm, "wText", stArg, iArg);
	XtManageChild(wText[i]);
	XtVaSetValues(wText[i], XtVaTypedArg, XmNforeground, XmRString, 
				  FgRegColor, strlen(FgRegColor)+1, NULL);
	XtVaSetValues(wText[i], XtVaTypedArg, XmNbackground, XmRString, 
				  BgRegColor, strlen(BgRegColor)+1, NULL);
	XtSetArg(stArg[3], XmNleftWidget, wText[i]);
	sLOffset = sLabelDistance;
	sPre = 1;
  }
  XtVaSetValues(wText[sCurButton], XtVaTypedArg, XmNbackground, XmRString, 
				BgSetColor, strlen(BgSetColor)+1, NULL);


  /*************************************************************************
   * Widget to display input character count 
   *************************************************************************/
  xmLen = XmStringCreateLocalized("");
  wlLength = XtVaCreateManagedWidget("wLabel", xmLabelGadgetClass, wForm,
									 XmNlabelString, xmLen,
									 XmNtopAttachment, XmATTACH_WIDGET,
									 XmNtopWidget, wForm,
									 XmNrightAttachment, XmATTACH_WIDGET,
									 XmNrightWidget, wForm,
									 XmNrightOffset, RIGHTOFFSET,
									 NULL);

  /*************************************************************************
   * Widget to display buffer of last 10 cut/paste  
   *************************************************************************/
  wPopup = XtVaCreatePopupShell("wPopup", xmDialogShellWidgetClass, wTop,
								XmNoverrideRedirect, True,
								NULL);

  wPopForm = XtVaCreateWidget("wPopForm", xmFormWidgetClass, wPopup,
							  NULL );

  iArg = 0;
  XtSetArg(stArg[iArg], XmNvisibleItemCount, LIST_COUNT); iArg++;
  wList = (Widget)XmCreateScrolledList(wPopForm, "wList", stArg, iArg);  

  XtManageChild(wList);

  XGetTextProperty(dpy, window, &Xcutbuf, XA_CUT_BUFFER0);
  XmListAddItemUnselected(wList, XmStringCreateLocalized(Xcutbuf.value), 1);

  /*************************************************************************
   * Callbacks 
   *************************************************************************/

  for (i = 0; i < BOX_COUNT; i++)
	XtAddCallback(wText[i], XmNfocusCallback, vPushed, (XtPointer)i);

  XtAddCallback(wList, XmNdefaultActionCallback, vDoubleClick, NULL);

  XtAddCallback(wTop, XmNdestroyCallback, (XtCallbackProc)vClose, NULL);
  XtAddCallback(wTop, XmNdestroyCallback, (XtCallbackProc)vClose, NULL);

  XtRealizeWidget(wTop);

  /*************************************************************************
   * Main loop 
   *************************************************************************/

  for (;;)
  {
	XtAppNextEvent(Xapp, &Xevent);

    if (Xevent.type == FocusOut && cFlag == 1)
	  vPopDown(NULL, NULL, wPopup);

	if (Xevent.type == PropertyNotify || 
		(Xevent.type == ButtonPress && Xevent.xbutton.button > Button1) ||
		(Xevent.type == KeyPress && Xevent.xkey.keycode == 36))
	{
	  if (Xevent.type & PropertyNotify)
	  {
		XGetTextProperty(dpy, window, &Xcutbuf, XA_CUT_BUFFER0);
		Xcutbuf.value[strcspn(Xcutbuf.value, "\n\0")] = 0;
		if (Xcutbuf.nitems)
		  vConvertInput((char *)Xcutbuf.value);

		xmLen = XmStringCreateLocalized((char *)Xcutbuf.value);
		if ((iPos = XmListItemPos(wList, xmLen)) != 0)
		{
		  XmListDeletePos(wList, iPos);
		  iItemCnt--;
		}
		XmListAddItemUnselected(wList, xmLen, 1);
		iItemCnt++;

		if (iItemCnt > 10)
		  XmListDeletePos(wList, iItemCnt--);

		free(Xcutbuf.value);
	  }
	  if (Xevent.type == KeyPress && Xevent.xkey.keycode == 36)
		vConvertInput(XmTextGetString(wText[sCurButton]));

	  if (Xevent.type == ButtonPress)
	  {
		if (Xevent.xbutton.button == Button3)
		{
		  XtManageChild(wPopForm);
		  XQueryPointer(dpy, window, &wRoot, &wChild, &iRx, &iRy, &iWx, &iWy, 
						&uiMask);
		  XtVaGetValues(wForm, XmNy, &iWy, NULL);
		
		  if (iWy < 170)
			iRy = iWy + XCON_HEIGHT;
		  else
			iRy = iWy - 168;

		  XtVaSetValues(wPopForm, XmNx, iRx-50, XmNy, iRy, NULL);
		  XtPopup(wPopup, XtGrabNone);
		  cFlag = 1;
		}
		if (Xevent.xbutton.button == Button2)
		  XmListDeleteAllItems(wList);
	  }
	}

	XtDispatchEvent(&Xevent);
  }
  return 0;
}


/*************************************************************************
 * Funciton to convert input
 *************************************************************************/
void vConvertInput(char *pcInput)
{
  XmString      xmStr;
  char *pcStop=NULL, *pcOut=(char *)malloc(sizeof(char)), caLen[11];
  
  sprintf(caLen, "%10d", strlen(pcInput));

  xmStr = XmStringCreate((String)caLen, caFont);
  XtVaSetValues(wlLength, XmNlabelString, xmStr, NULL);

  switch(sCurButton)
  {
    case 0:
	  ulVal = strtoul(pcInput, &pcStop, 16);
	  if (ulVal && !*pcStop)
	  {
		XmTextSetString(wText[0], pcInput);
			
		if (ulVal < ~0UL)
		{
		  vHEXToDEC(ulVal, &pcOut);
		  XmTextSetString(wText[1], pcOut);
		}
		else
		  XmTextSetString(wText[1], "Overflow");
			  
		vHEXToABC(pcInput, &pcOut);
		XmTextSetString(wText[2], pcOut);
	  }
	  break;
    case 1:
	  ulVal = strtoul(pcInput, &pcStop, 10);
	  if (ulVal && !*pcStop && ulVal < ~0UL)
	  {
		if (ulVal < ~0UL)
		{
		  vDECToHEX(ulVal, &pcOut);
		  XmTextSetString(wText[0], pcOut);
		  pcStop = (char *)malloc(strlen(pcOut)+1);
		  sprintf(pcStop, "%s", pcOut);
		}			
		else
		  XmTextSetString(wText[0], "Overflow");
			  
		XmTextSetString(wText[1], pcInput);
			
		if (ulVal < ~0UL)
		{
		  vHEXToABC(pcStop, &pcOut);
		  XmTextSetString(wText[2], pcOut);
		}
		else
		  XmTextSetString(wText[2], "Not applicable");
		free(pcStop);
	  }
	  break;
    default:
	  vABCToHEX(pcInput, &pcOut);
	  XmTextSetString(wText[0], pcOut);

	  vABCToDEC(pcInput, &pcOut);
	  XmTextSetString(wText[1], pcOut);

	  XmTextSetString(wText[2], pcInput);
	  break;
  }
  free(pcOut);
}

/*************************************************************************
 * Funciton to convert from text to hex
 *************************************************************************/

void vABCToHEX(char *pcData, char **ppcOut)
{
  ulong ulIdx, ulLen = strlen(pcData);
  
  *ppcOut = (char *)realloc(*ppcOut, sizeof(char) * ulLen * 3 + 1);
  for (ulIdx = 0; ulIdx < ulLen; ulIdx++)
  {
	sprintf(&(*ppcOut)[ulIdx*3], "%02x", (ushort)pcData[ulIdx]);
	if (ulIdx < ulLen - 1)
	  (*ppcOut)[ulIdx*3+2] = '-';
  }
  return;
}

/*************************************************************************
 * Funciton to convert from dec to hex
 *************************************************************************/

void vDECToHEX(ulong ulData, char **ppcOut)
{
  *ppcOut = (char *)realloc(*ppcOut, sizeof(char) * 9);
  sprintf(*ppcOut, "%lx", ulData);
  return;
}

/*************************************************************************
 * Funciton to convert from text to dec
 *************************************************************************/

void vABCToDEC(char *pcData, char **ppcOut)
{
  ulong ulIdx, ulLen = strlen(pcData);

  *ppcOut = (char *)realloc(*ppcOut, sizeof(char) * ulLen * 4 + 1);
  for (ulIdx = 0; ulIdx < ulLen; ulIdx++)
  {
	sprintf(&(*ppcOut)[ulIdx*4], "%03d", pcData[ulIdx]);
	if (ulIdx < ulLen - 1)
	  (*ppcOut)[ulIdx*4+3] = '-';
  }
  return;
}

/*************************************************************************
 * Funciton to convert from hex to dec
 *************************************************************************/

void vHEXToDEC(ulong ulData, char **ppcOut)
{
  *ppcOut = (char *)realloc(*ppcOut, sizeof(char) * 11);
  sprintf(*ppcOut, "%lu", ulData);
  return;
}

/*************************************************************************
 * Function to convert from hex to text
 *************************************************************************/

void vHEXToABC(char *pcData, char **ppcOut)
{
  ulong ulPos, ulLen = strlen(pcData);
  long  lIdx;
  char  *pcTemp = (char *)calloc(ulLen + 2, 1), caChar[3] = "  \0";

  ulPos = (ulLen + 1) / 2 + 1;
  *ppcOut = (char *)realloc(*ppcOut, sizeof(char) * ulPos);
  (*ppcOut)[ulPos-1] = 0;
  if (ulLen % 2)
	strncpy(&pcTemp[1], pcData, ulLen++);
  else
	strcpy(pcTemp, pcData);
  ulLen -= 2;
  for (lIdx = ulLen; lIdx >= 0; lIdx-=2)
  {
	memcpy(caChar, &pcTemp[lIdx], 2);
	(*ppcOut)[lIdx / 2] = (char)strtoul(caChar, NULL, 16);
#if 0
	if (!isprint((*ppcOut)[lIdx / 2]))
	  (*ppcOut)[lIdx / 2] = '.';
#endif
  }
  return;
}

/*************************************************************************
 * Callback for button push action
 *************************************************************************/

void vPushed(Widget widget, XtPointer client_data, XtPointer call_data)
{
  XtVaSetValues(wText[sCurButton], XtVaTypedArg, XmNforeground, XmRString, 
				FgRegColor, strlen(FgRegColor)+1, XtVaTypedArg, XmNbackground,
				XmRString, BgRegColor, strlen(BgRegColor)+1, NULL);
  sCurButton = (int)client_data;
  XtVaSetValues(wText[sCurButton], XtVaTypedArg, XmNforeground, XmRString, 
				FgSetColor, strlen(FgSetColor)+1, XtVaTypedArg, XmNbackground,
				XmRString, BgSetColor, strlen(BgSetColor)+1, NULL);
}

/*************************************************************************
 * Callback for default action
 *************************************************************************/
void vDoubleClick(Widget widget, XtPointer client_data, XtPointer call_data)
{
  XmString *pxmSel;

  XtVaGetValues(wList, XmNselectedItems, &pxmSel, NULL);
  XmStringGetLtoR(pxmSel[0], XmFONTLIST_DEFAULT_TAG, (char **)&Xcutbuf.value);
  Xcutbuf.nitems = strlen(Xcutbuf.value);
  XSetTextProperty(dpy, window, &Xcutbuf, XA_CUT_BUFFER0);
  vPopDown(widget, client_data, call_data);
}

/*************************************************************************
 * Callback for closing popup list
 *************************************************************************/

void vPopDown(Widget widget, XtPointer client_data, XtPointer call_data)
{
  XtUnmanageChild(wPopForm);
  XtPopdown(wPopup);
  cFlag = 0;
}

/*************************************************************************
 * Callback for closing window
 *************************************************************************/

void vClose(void)
{
  free(pcOut);
  exit(0);
}
